{*******************************************************}
{                                                       }
{       Alex Ghost Library                              }
{                                                       }
{       Copyright (c) 1999,2000 Alexey Popov            }
{                                                       }
{*******************************************************}

{$I AG.INC}

unit agAppInst;

interface

uses Windows, SysUtils, Classes, Forms;

const
  E_INVALID_CONTROL_FILE = 'Could not open or create "%0:s"';
  E_LOCK_TIMEOUT = 'Could not lock "%0:s"';
  E_WRITE_ERROR = 'Could not write to "%0:s"';
  E_CONTROL_FILE_CHANGE = 'ControlFile can not be changed when Active = True';
  E_USERNAME_CHANGE = 'UserName can not be changed when Active = True';
  E_COMMENT_CHANGE = 'Comment can not be changed when Active = True';
  E_INTERNAL_ERROR = 'Internal error in TagAppInstance';

  LOCK_OFFSET   = $80000000;
  LOCK_TIMEOUT  = 5.0; { sec }

type

  EAppInstanceError = class(Exception);

  TagAppInstance = class(TObject)
  private
    Recno: integer;
    hControlFile: THandle;
    FInstance: pointer;
    FControlFile: string;
    FInstances: TList;
    FEnabled: boolean;
    FSizeOfInstance: integer;
    procedure OpenControlFile;
    procedure CloseControlFile;
    procedure LockNextFreeRecord;
    function  LockRecord (nRecno: Integer): Boolean;
    procedure UnlockRecord (nRecno: Integer);
    procedure FindInstance;
    procedure LockControlFile;
    procedure UnlockControlFile;
    procedure SetActive(Value: Boolean);
    function  GetActive: Boolean;
    procedure SetControlFile(Value: string);
    procedure SetEnabled(Value: Boolean);
    procedure InternalRefresh;
    procedure ClearInstances;
  public
    constructor Create(SizeOfInstance: integer);
    destructor Destroy; override;
    procedure Open;
    procedure Close;
    property Instances: TList read FInstances;
    function Fail: boolean;
    procedure Refresh;
    property Active: Boolean read GetActive write SetActive;
    property ControlFile: string read FControlFile write SetControlFile;
    property Enabled: boolean read FEnabled write SetEnabled;
    property SizeOfInstance: integer read FSizeOfInstance;
    property Instance: pointer read FInstance;
    procedure WriteInstance;
  end;

implementation

constructor TagAppInstance.Create(SizeOfInstance: integer);
begin
  inherited Create;
  FSizeOfInstance:=SizeOfInstance;
  GetMem(FInstance,FSizeOfInstance);
  FillChar(FInstance^,FSizeOfInstance,0);
  hControlFile:=INVALID_HANDLE_VALUE;
  FInstances:=TList.Create;
  FEnabled:=true;
end;

destructor TagAppInstance.Destroy;
begin
  SetActive(false);
  ClearInstances;
  FInstances.Free;
  FreeMem(FInstance);
  inherited Destroy;
end;

procedure TagAppInstance.ClearInstances;
var
  i: integer;
begin
  for i:=0 to FInstances.Count-1 do FreeMem(FInstances[i]);
  FInstances.Clear;
end;

procedure TagAppInstance.Open;
begin
  SetActive(true);
end;

procedure TagAppInstance.Close;
begin
  SetActive(false);
end;

procedure TagAppInstance.InternalRefresh;
var
  Inst: pointer;
  nRecno: Integer;
begin
  if Active then begin
    ClearInstances;
    FileSeek(hControlFile, 0, 0);
    nRecno:=0;
    GetMem(Inst,FSizeOfInstance);
    try
      while FileRead(hControlFile,Inst^,FSizeOfInstance) = FSizeOfInstance do begin
        Inc(nRecno);
        if LockRecord(nRecno) then
          UnLockRecord(nRecno)
        else begin
          FInstances.Add(Inst);
          GetMem(Inst,FSizeOfInstance);
        end;
      end;
    finally
      FreeMem(Inst);
    end;
  end;
end;

procedure TagAppInstance.SetActive(Value: Boolean);
begin
  if Active <> Value then begin
    if Value then begin
      OpenControlFile;
      if FEnabled then begin
        LockNextFreeRecord;
        WriteInstance;
      end;
      InternalRefresh;
    end else begin
      if FEnabled then UnlockRecord (Recno);
      CloseControlFile;
      ClearInstances;
    end;
  end;
end;

function TagAppInstance.GetActive: Boolean;
begin
  Result:=(hControlFile <> INVALID_HANDLE_VALUE);
end;

procedure TagAppInstance.SetControlFile(Value: string);
begin
  if Value <> FControlFile then begin
    if not Active then
      FControlFile:=Value
    else
      raise EAppInstanceError.Create(E_CONTROL_FILE_CHANGE);
  end;
end;

procedure TagAppInstance.OpenControlFile;
begin
  hControlFile:=CreateFile(PChar(FControlFile),GENERIC_READ or GENERIC_WRITE,
    FILE_SHARE_READ or FILE_SHARE_WRITE,nil,OPEN_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
  if hControlFile = INVALID_HANDLE_VALUE then
    raise EAppInstanceError.CreateFmt(E_INVALID_CONTROL_FILE,[FControlFile]);
end;

procedure TagAppInstance.CloseControlFile;
begin
  CloseHandle(hControlFile);
  hControlFile:=INVALID_HANDLE_VALUE;
  if FInstances.Count <= 1 then DeleteFile(FControlFile);
end;

procedure TagAppInstance.LockNextFreeRecord;
var
  Locked,NewRecord: boolean;
  Inst: pointer;
begin
  FileSeek(hControlFile,0,0);
  Recno:=0;
  NewRecord:=false;
  Locked:=false;
  GetMem(Inst,FSizeOfInstance);
  try
    repeat
      if FileRead(hControlFile,Inst^,FSizeOfInstance) = FSizeOfInstance then begin
        Inc(Recno);
        Locked:=LockRecord(Recno);
      end else begin
        NewRecord:=true;
      end;
    until Locked or NewRecord;
  finally
    FreeMem(Inst);
  end;
  if NewRecord then begin
    LockControlFile;
    try
      Recno:=(FileSeek(hControlFile,0,2)+1) div FSizeOfInstance + 1;
      if not LockRecord(Recno) then
        raise EAppInstanceError.Create(E_INTERNAL_ERROR);
      WriteInstance;
    finally
      UnlockControlFile;
    end;
  end else
    WriteInstance;
end;

function TagAppInstance.LockRecord(nRecno: Integer): Boolean;
begin
  Result:=LockFile(hControlFile,LOCK_OFFSET+nRecno,0,1,0);
end;

procedure TagAppInstance.UnlockRecord(nRecno: Integer);
begin
  UnLockFile(hControlFile,LOCK_OFFSET+nRecno,0,1,0);
end;

procedure TagAppInstance.LockControlFile;
var
  TimeOut: TDateTime;
begin
  TimeOut:=Now + LOCK_TIMEOUT/24/60/60;
  while not LockRecord(0) do begin
    Application.ProcessMessages;
    if Now > TimeOut then
      raise EAppInstanceError.CreateFmt(E_LOCK_TIMEOUT,[FControlFile]);
  end;
end;

procedure TagAppInstance.UnLockControlFile;
begin
  UnlockRecord(0);
end;

procedure TagAppInstance.FindInstance;
begin
  FileSeek(hControlFile,(Recno-1)*FSizeOfInstance,0);
end;

procedure TagAppInstance.WriteInstance;
begin
  if FEnabled then begin
    FindInstance;
    if FileWrite(hControlFile,FInstance^,FSizeOfInstance) < FSizeOfInstance then
      raise EAppInstanceError.CreateFmt(E_WRITE_ERROR,[FControlFile]);
    FlushFileBuffers(hControlFile);
  end;
end;

function TagAppInstance.Fail: boolean;
var
  Inst: pointer;
begin
  Result:=true;
  if hControlFile <> INVALID_HANDLE_VALUE then begin
    FileSeek(hControlFile,0,0);
    GetMem(Inst,FSizeOfInstance);
    try
      if FileRead(hControlFile,Inst^,FSizeOfInstance) = FSizeOfInstance then
        Result:=false;
    finally
      FreeMem(Inst);
    end;
  end;
end;

procedure TagAppInstance.Refresh;
begin
  if Fail then begin
    CloseControlFile;
    ClearInstances;
    SetActive(true);
  end else
    InternalRefresh;
end;

procedure TagAppInstance.SetEnabled(Value: Boolean);
var
  act: boolean;
begin
  if Value <> FEnabled then begin
    act:=GetActive;
    SetActive(false);
    FEnabled:=Value;
    SetActive(act);
  end;
end;

end.
