{*******************************************************}
{                                                       }
{       Alex Ghost Library                              }
{                                                       }
{       Copyright (c) 1999,2000 Alexey Popov            }
{                                                       }
{*******************************************************}

{$I AG.INC}

unit agFoldMon;

interface

uses Classes;

type
  TFileChangeFilter = (fnFileName, fnDirName, fnAttributes, fnSize,
    fnLastWrite, fnLastAccess);
  TFileChangeFilters = set of TFileChangeFilter;
  TScanMode = (smEconomy, smIntensive);

  TNotifyThread = class(TThread)
  private
    FFolderName: string;
    FFilter: TFileChangeFilters;
    FInterval,FDelay,FScanDelay: Cardinal;
    FOnChange: TNotifyEvent;
    procedure CallOnChange;
  protected
    procedure DoChange; virtual;
    procedure Execute; override;
  public
    constructor Create(const FolderName: string; Filter: TFileChangeFilters;
      Interval,Delay: cardinal; ScanMode: TScanMode);
    destructor Destroy; override;
    property Terminated;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  TagFolderMonitor = class(TComponent)
  private
    FNotifyThread: TNotifyThread;
    FFilter: TFileChangeFilters;
    FDelay,FInterval: Cardinal;
    FFolderName: string;
    FActive,FNeedActive: Boolean;
    FPriority: TThreadPriority;
    FScanMode: TScanMode;
    FOnChange: TNotifyEvent;
    procedure SetActive(Value: Boolean);
    procedure SetFilter(Value: TFileChangeFilters);
    procedure SetFolderName(const Value: string);
    procedure SetInterval(Value: cardinal);
    procedure SetDelay(Value: cardinal);
    procedure SetPriority(Value: TThreadPriority);
    procedure SetScanMode(Value: TScanMode);
    procedure ThreadNotification(Sender: TObject);
  protected
    procedure Loaded; override;
    procedure FreeNotifyThread;
    procedure ResetNotifyThread(Activate: Boolean); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property Active: Boolean read FActive write SetActive default false;
    property Interval: Cardinal read FInterval write SetInterval default 0;
    property DelayTime: Cardinal read FDelay write SetDelay default 0;
    property Filter: TFileChangeFilters read FFilter write SetFilter
      default [fnFileName, fnDirName, fnSize, fnLastWrite];
    property FolderName: string read FFolderName write SetFolderName;
    property ThreadPriority: TThreadPriority read FPriority write SetPriority
      default tpNormal;
    property ScanMode: TScanMode read FScanMode write SetScanMode default smEconomy;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

implementation

uses Windows, SysUtils, agFileUtils, agFileFind;

type
  PFileRec = ^TFileRec;
  TFileRec = record
    FileName: string;
    IsDir: boolean;
    Size: integer;
    Attr: integer;
    LastWrite,LastAccess: TFileTime; 
  end;

{ TNotifyThread }

constructor TNotifyThread.Create(const FolderName: string;
  Filter: TFileChangeFilters; Interval,Delay: cardinal; ScanMode: TScanMode);
begin
  FFolderName:=FolderName;
  FFilter:=Filter;
  FDelay:=Delay;
  FInterval:=Interval;
  if ScanMode = smEconomy then FScanDelay:=1 else FScanDelay:=0;
  inherited Create(true);
end;

destructor TNotifyThread.Destroy;
begin
  FOnChange:=nil;
  inherited Destroy;
end;

procedure TNotifyThread.CallOnChange;
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

procedure TNotifyThread.DoChange;
begin
  if Assigned(FOnChange) then Synchronize(CallOnChange);
end;

procedure TNotifyThread.Execute;
var
  Files: TList;
  NumFiles,NumDirs: cardinal;

  procedure ClearFiles;
  var
    i: integer;
  begin
    for i:=0 to Files.Count-1 do dispose(PFileRec(Files[i]));
    Files.Clear;
  end;

  function IsDirAttr(FileAttr: Integer): boolean;
  begin
    Result:=((FileAttr and faDirectory) = faDirectory);
  end;

  function IndexOfFile(FileName: string): integer;
  var
    i: integer;
  begin
    Result:=-1;
    for i:=0 to Files.Count-1 do
      if CompareText(PFileRec(Files[i])^.FileName,FileName) = 0 then begin
        Result:=i;
        break;
      end;
  end;

  procedure StoreCurrentState;
  var
    i: integer;
    sr: TSearchRecEx;
    F: PFileRec;
  begin
    ClearFiles;
    NumFiles:=0; NumDirs:=0;
    i:=FindFirstEx(FFolderName+'*.*',0,faVolumeID,sr);
    try
      while i = 0 do begin
        if Terminated then exit;
        new(F);
        with F^ do begin
          FileName:=sr.Name;
          IsDir:=IsDirAttr(sr.Attr);
          if IsDir then inc(NumDirs) else inc(NumFiles);
          Size:=sr.Size;
          Attr:=sr.Attr;
          LastWrite:=sr.FindData.ftLastWriteTime;
          LastAccess:=sr.FindData.ftLastAccessTime;
        end;
        Files.Add(F);
        i:=FindNextEx(sr);
      end;
    finally
      FindCloseEx(sr);
    end;
  end;

  function StateChanged: boolean;
  var
    i,j: integer;
    nf,nd: cardinal;
    sr: TSearchRecEx;
  begin
    Result:=false;
    nf:=0; nd:=0;
    i:=FindFirstEx(FFolderName+'*.*',0,faVolumeID,sr);
    try
      while i = 0 do begin
        if Terminated then exit;
        Sleep(FScanDelay);
        j:=IndexOfFile(sr.Name);
        if j <> -1 then
          if PFileRec(Files[j])^.IsDir <> IsDirAttr(sr.Attr) then j:=-1;
        if j = -1 then begin
          // check for new
          if (fnFileName in FFilter) and not IsDirAttr(sr.Attr) then begin
            Result:=true;
            exit;
          end;
          if (fnDirName in FFilter) and IsDirAttr(sr.Attr) then begin
            Result:=true;
            exit;
          end;
        end else
          with PFileRec(Files[j])^ do
            if IsDir then
              inc(nd)
            else begin
              inc(nf);
              // check for changed
              if fnSize in FFilter then
                if Size <> sr.Size then begin
                  Result:=true;
                  exit;
                end;
              if fnAttributes in FFilter then
                if Attr <> sr.Attr then begin
                  Result:=true;
                  exit;
                end;
              if fnLastWrite in FFilter then
                if not IsEqualFileTime(LastWrite,sr.FindData.ftLastWriteTime) then begin
                  Result:=true;
                  exit;
                end;
              if fnLastAccess in FFilter then
                if not IsEqualFileTime(LastAccess,sr.FindData.ftLastAccessTime) then begin
                  Result:=true;
                  exit;
                end;
            end;
        i:=FindNextEx(sr);
      end;
    finally
      FindCloseEx(sr);
    end;
    // check for deleted
    if not Result then begin
      if (fnFileName in FFilter) and (nf < NumFiles) then Result:=true;
      if (fnDirName in FFilter) and (nd < NumDirs) then Result:=true;
    end;
  end;

  procedure Pause(ADelay: cardinal);
  var
    FirstTickCount, Now: Longint;
  begin
    if ADelay = 0 then exit;
    FirstTickCount:=GetTickCount;
    repeat
      Sleep(1);
      if Terminated then exit;
      Now:=GetTickCount;
    until (Now - FirstTickCount >= ADelay) or (Now < FirstTickCount);
  end;

begin
  Files:=TList.Create;
  try
    StoreCurrentState;
    while not Terminated do begin
      if StateChanged then begin
        Pause(FDelay);
        if Terminated then exit;
        DoChange;
        StoreCurrentState;
      end;
      Pause(FInterval);
    end;
  finally
    ClearFiles;
    Files.Free;
  end;
end;

{ TagFolderMonitor }

constructor TagFolderMonitor.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FFilter:=[fnFileName,fnDirName,fnSize,fnLastWrite];
  FPriority:=tpNormal;
end;

destructor TagFolderMonitor.Destroy;
begin
  FreeNotifyThread;
  inherited Destroy;
end;

procedure TagFolderMonitor.Loaded;
begin
  inherited Loaded;
  if FNeedActive then ResetNotifyThread(true);
end;

procedure TagFolderMonitor.SetActive(Value: Boolean);
begin
  if FActive <> Value then begin
    FActive:=Value;
    ResetNotifyThread(FActive);
  end;
end;

procedure TagFolderMonitor.SetFilter(Value: TFileChangeFilters);
begin
  if FFilter <> Value then begin
    FFilter := Value;
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.SetFolderName(const Value: string);
begin
  if FFolderName <> Value then begin
    FFolderName:=NormalDir(Value);
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.SetInterval(Value: cardinal);
begin
  if FInterval <> Value then begin
    FInterval:=Value;
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.SetDelay(Value: cardinal);
begin
  if FDelay <> Value then begin
    FDelay:=Value;
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.SetPriority(Value: TThreadPriority);
begin
  if FPriority <> Value then begin
    FPriority:=Value;
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.SetScanMode(Value: TScanMode);
begin
  if FScanMode <> Value then begin
    FScanMode:=Value;
    ResetNotifyThread(Active);
  end;
end;

procedure TagFolderMonitor.FreeNotifyThread;
begin
  if FNotifyThread <> nil then
    with FNotifyThread do begin
      OnChange:=nil;
      FreeOnTerminate:=true;
      Terminate;
    end;
  FNotifyThread:=nil;
end;

procedure TagFolderMonitor.ResetNotifyThread(Activate: Boolean);
begin
  if csDesigning in ComponentState then exit;
  if (csReading in ComponentState) then
    FNeedActive:=Activate
  else begin
    FreeNotifyThread;
    if Activate and DirExists(FFolderName) then begin
      FNotifyThread:=TNotifyThread.Create(FFolderName,FFilter,FInterval,FDelay,FScanMode);
      with FNotifyThread do begin
        OnChange:=ThreadNotification;
        Priority:=FPriority;
        Resume;
      end;
    end;
  end;
end;

procedure TagFolderMonitor.ThreadNotification(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Self);
end;

end.
