{*******************************************************}
{                                                       }
{       Alex Ghost Library                              }
{                                                       }
{       Copyright (c) 1999,2000 Alexey Popov            }
{                                                       }
{*******************************************************}

{$I AG.INC}

unit agGraphUtils;

interface

uses Windows, agUtils;

function PointInRect(const P: TPoint; const R: TRect): Boolean;


type
  TFillDirection = (fdTopToBottom, fdBottomToTop, fdLeftToRight, fdRightToLeft);

// Gradient filling routine
procedure GradientFillRect(DC: HDC; ARect: TRect; StartColor, EndColor: TColor;
  Direction: TFillDirection; Colors: Word);

  
type
  TAlignment = (taLeftJustify, taRightJustify, taCenter);
  TVertAlignment = (vaTopJustify, vaCenter, vaBottomJustify);

procedure DrawTextWrap(DC: HDC; const Text: string; ClipRect: TRect;
  HAlign: TAlignment; VAlign: TVertAlignment; Trans: boolean);

  
implementation

uses agArithm, agStrUtils;

function PointInRect(const P: TPoint; const R: TRect): Boolean;
begin
  with R do
    Result := (Left <= P.X) and (Top <= P.Y) and (Right >= P.X) and (Bottom >= P.Y);
end;

procedure GradientFillRect(DC: HDC; ARect: TRect; StartColor, EndColor: TColor;
  Direction: TFillDirection; Colors: Word);
var
  StartRGB: array[0..2] of Byte;    { Start RGB values }
  RGBDelta: array[0..2] of Integer; { Difference between start and end RGB values }
  ColorBand: TRect;                 { Color band rectangular coordinates }
  I: Integer;
  Delta: real;
  Brush: HBrush;
begin
  if IsRectEmpty(ARect) then Exit;
  if Colors < 2 then begin
    Brush := CreateSolidBrush(ColorToRGB(StartColor));
    FillRect(DC, ARect, Brush);
    DeleteObject(Brush);
    Exit;
  end;
  StartColor := ColorToRGB(StartColor);
  EndColor := ColorToRGB(EndColor);
  case Direction of
    fdTopToBottom, fdLeftToRight: begin
      { Set the Red, Green and Blue colors }
      StartRGB[0] := GetRValue(StartColor);
      StartRGB[1] := GetGValue(StartColor);
      StartRGB[2] := GetBValue(StartColor);
      { Calculate the difference between begin and end RGB values }
      RGBDelta[0] := GetRValue(EndColor) - StartRGB[0];
      RGBDelta[1] := GetGValue(EndColor) - StartRGB[1];
      RGBDelta[2] := GetBValue(EndColor) - StartRGB[2];
    end;
    fdBottomToTop, fdRightToLeft: begin
      { Set the Red, Green and Blue colors }
      { Reverse of TopToBottom and LeftToRight directions }
      StartRGB[0] := GetRValue(EndColor);
      StartRGB[1] := GetGValue(EndColor);
      StartRGB[2] := GetBValue(EndColor);
      { Calculate the difference between begin and end RGB values }
      { Reverse of TopToBottom and LeftToRight directions }
      RGBDelta[0] := GetRValue(StartColor) - StartRGB[0];
      RGBDelta[1] := GetGValue(StartColor) - StartRGB[1];
      RGBDelta[2] := GetBValue(StartColor) - StartRGB[2];
    end;
  end;
  { Calculate the color band's coordinates }
  ColorBand := ARect;
  if Direction in [fdTopToBottom, fdBottomToTop] then begin
    Colors := MaxInt([2, MinInt([Colors, HeightOf(ARect)])]);
    Delta := HeightOf(ARect) / Colors;
  end
  else begin
    Colors := MaxInt([2, MinInt([Colors, WidthOf(ARect)])]);
    Delta := WidthOf(ARect) / Colors;
  end;
  { Perform the fill }
  if Delta > 0 then begin
    for I := 0 to Colors - 1 do begin
      case Direction of
        { Calculate the color band's top and bottom coordinates }
        fdTopToBottom, fdBottomToTop: begin
          ColorBand.Top := Round(ARect.Top + I * Delta);
          ColorBand.Bottom := Round(ColorBand.Top + Delta) + 1;
        end;
        { Calculate the color band's left and right coordinates }
        fdLeftToRight, fdRightToLeft: begin
          ColorBand.Left := Round(ARect.Left + I * Delta);
          ColorBand.Right := Round(ColorBand.Left + Delta) + 1;
        end;
      end;
      { Calculate the color band's color }
      Brush := CreateSolidBrush(RGB(
        StartRGB[0] + MulDiv(I, RGBDelta[0], Colors),
        StartRGB[1] + MulDiv(I, RGBDelta[1], Colors),
        StartRGB[2] + MulDiv(I, RGBDelta[2], Colors)));
      FillRect(DC, ColorBand, Brush);
      DeleteObject(Brush);
    end;
  end;
end;

procedure DrawTextWrap(DC: HDC; const Text: string; ClipRect: TRect;
  HAlign: TAlignment; VAlign: TVertAlignment; Trans: boolean);
const
  HorzAl: array [TAlignment] of integer = (DT_LEFT, DT_RIGHT, DT_CENTER);
var
  s: string;
  dy: integer;
begin
  s:=TextToLinesDC(DC,Text,WidthOf(ClipRect));
  dy:=(HeightOf(ClipRect)-(TextSizeDC(DC,s).cy*WordCount(s,[#13]))) div 2;
  if dy < 0 then dy:=0;
  case VAlign of
    vaTopJustify:;
    vaCenter:
      begin
        inc(ClipRect.Top,dy); dec(ClipRect.Bottom,dy);
      end;
    vaBottomJustify:
      inc(ClipRect.Top,dy*2);
  end;
  if Trans then
    SetBkMode(DC,TRANSPARENT)
  else
    SetBkMode(DC,OPAQUE);
  DrawText(DC,PChar(s),length(s),ClipRect,HorzAl[HAlign]);
end;

end.
