{*******************************************************}
{                                                       }
{       Alex Ghost Library                              }
{                                                       }
{       Copyright (c) 1999,2000 Alexey Popov            }
{                                                       }
{*******************************************************}

{$I AG.INC}

unit agIniFile;

interface

uses Classes, SysUtils;

type
  TEraseSectionCallback = function(const sectionName: string;
    const sl1,sl2: TStringList): boolean of object;

  TSectionList = class(TStringList)
  private
    FPrevIndex: integer;
  public
    constructor Create;
    function IndexOf(const S: AnsiString): Integer; override;
    function EraseDuplicates(callBackProc: TEraseSectionCallback): boolean;
    function GetSectionItems(Index: Integer): TStringList;
    property SectionItems[Index: Integer]: TStringList read GetSectionItems;
  end;

  TagIniFile = class(TObject)
  private
    FFileName: String;
    FFileHandle: integer;
    FHasChanged: Boolean;
    FSectionList: TSectionList;
    FPrevSectionIndex: Integer;
    FReadOnly: boolean;
    FEraseSectionCallback: TEraseSectionCallback;
    FFlagDropCommentLines,           // set false to keep lines starting with ';'
    FFlagFilterOutInvalid,           // set false to keep lines without '='
    FFlagDropWhiteSpace,             // set false to keep white space around '='
    FFlagDropApostrophes,            // set false to keep apostrophes around key
    FFlagTrimRight: Boolean;         // set false to keep white space at end of line

    function FindItemIndex(const aSection,aKey: String; CreateNew: Boolean;
      var FoundStringList: TStringList): integer;
    procedure SetFileName(const aName: String);
    procedure ClearSectionList;
  protected
    procedure AppendFromFile(const aName: String);
  public
    constructor Create(const FileName: String; ReadOnly: boolean);
    destructor Destroy; override;
    function ReadString(const aSection, aKey, aDefault: string): string; virtual;
    procedure WriteString(const aSection, aKey, aValue: String); virtual;
    function ReadInteger(const aSection, aKey: string; aDefault: Longint): Longint; virtual;
    procedure WriteInteger(const aSection, aKey: string; aValue: Longint); virtual;
    function ReadBool(const aSection, aKey: string; aDefault: Boolean): Boolean; virtual;
    procedure WriteBool(const aSection, aKey: string; aValue: Boolean); virtual;
    procedure ReadSection(const aSection: string; aStrings: TStrings); virtual;
    procedure ReadSections(Strings: TStrings); virtual;
    procedure ReadSectionValues(const aSection: string; aStrings: TStrings); virtual;
    procedure EraseSection(const aSection: string); virtual;
    procedure DeleteKey(const aSection, aKey: string); virtual;
    procedure ReadAll(aStrings: TStrings); virtual;
    procedure FlushFile; virtual;
    property SectionNames: TSectionList read FSectionList;
    property FileName: string read FFileName write SetFileName;
    property EraseSectionCallback: TEraseSectionCallback read FEraseSectionCallback
      write FEraseSectionCallback;
    property FlagDropCommentLines: Boolean read FFlagDropCommentLines
      write FFlagDropCommentLines;
    property FlagFilterOutInvalid: Boolean read FFlagFilterOutInvalid
      write FFlagFilterOutInvalid;
    property FlagDropWhiteSpace: Boolean read FFlagDropWhiteSpace write FFlagDropWhiteSpace;
    property FlagDropApostrophes: Boolean read FFlagDropApostrophes write FFlagDropApostrophes;
    property FlagTrimRight: Boolean read FFlagTrimRight write FFlagTrimRight;
  end;

implementation

uses Windows, agFileUtils;

function max(a,b: Integer): Integer;
begin
  if a > b then result:=a else result:=b;
end;

constructor TSectionList.Create;
begin
  inherited Create;
  FPrevIndex:=0;
end;

function TSectionList.GetSectionItems(Index: Integer): TStringList;
begin
  result:=TStringList(Objects[Index]);
end;

function TSectionList.EraseDuplicates(callBackProc: TEraseSectionCallback): Boolean;
var
  slDuplicateTracking: TStringList;
  idxToDelete,ixLow,ixHigh,ix: integer;

  procedure SwapInt(var a,b: Integer);
  var
    c : Integer;
  begin
    c := a; a := b; b := c;
  end;
  
begin
  Result:=False;

  if count > 1 then
  begin
    slDuplicateTracking:=TStringList.Create;
    slDuplicateTracking.Assign(Self);
    for ix:=0 to slDuplicateTracking.count-1 do
      slDuplicateTracking.Objects[ix]:=Pointer(ix);
    slDuplicateTracking.Sort;
    ixLow:=0;
    for ix:=1 to slDuplicateTracking.count-1 do
    begin
      if (AnsiCompareText(slDuplicateTracking.Strings[ixLow],
        slDuplicateTracking.Strings[ix]) <> 0)
      then ixLow:=ix
      else begin
        ixHigh := ix;
        if Integer(slDuplicateTracking.Objects[ixLow]) >
           Integer(slDuplicateTracking.Objects[ixHigh]) then swapInt(ixHigh,ixLow);

        if assigned(callBackProc) then begin
          if NOT CallBackProc(slDuplicateTracking.Strings[ix],
            SectionItems[Integer(slDuplicateTracking.Objects[ixLow])],
            SectionItems[Integer(slDuplicateTracking.Objects[ixHigh])])
            then swapInt(ixHigh,ixLow);
        end;
        idxToDelete := Integer(slDuplicateTracking.Objects[ixHigh]);

        SectionItems[idxToDelete].Clear;
        Objects[idxToDelete] := nil;
        Result := True; { list had been changed }
      end;
    end;

    ix := 0;
    while ix < count do
      if Objects[ix] = NIL then delete(ix) else inc(ix);
    slDuplicateTracking.free;
  end;

end;

function TSectionList.IndexOf(const S: AnsiString): Integer;
var
  IX,LastIX: Integer;
begin
  Result := -1;
  If Count = 0 then Exit;

  LastIX := FPrevIndex;
  for IX := LastIX to count-1 do
    if (AnsiCompareText(Get(IX), S) = 0) then begin
      Result     := IX;
      FPrevIndex := IX;
      exit;
    end;
  for IX := 0 to LastIX-1 do
    if (AnsiCompareText(Get(IX), S) = 0) then begin
      Result     := IX;
      FPrevIndex := IX;
      exit;
    end;
end;

constructor TagIniFile.Create(const FileName: string; ReadOnly: boolean);
begin
  FSectionList:=TSectionList.Create;
  FFlagDropCommentLines:=true;
  FFlagFilterOutInvalid:=true;
  FFlagDropWhiteSpace:=true;
  FFlagDropApostrophes:=true;
  FFlagTrimRight:=true;
  FFileName:='';
  FPrevSectionIndex:=0;
  FEraseSectionCallback:=nil;
  FFileHandle:=0;
  FReadOnly:=ReadOnly;
  SetFileName(FileName);
end;

destructor TagIniFile.Destroy;
begin
  FlushFile;
  if FFileHandle > 0 then agFileClose(FFileHandle);
  ClearSectionList;
  FSectionList.Free;
  inherited Destroy;
end;

procedure TagIniFile.ClearSectionList;
var
  ixSections: Integer;
begin
  with FSectionList do begin
    for ixSections := 0 to count -1 do
      SectionItems[ixSections].Free;
    clear;
  end;
end;

procedure TagIniFile.AppendFromFile(const aName : String);
var
  CurrStringList: TStringList;
  CurrSectionName,OneLine: String;
  LL,LastPos,EqPos: integer;
  nospace: boolean;
  openmode: DWORD;
  sharemode,createmode: integer;
begin
  CurrStringList := NIL;
  FPrevSectionIndex := 0;
  if FFileHandle > 0 then begin
    agFileClose(FFileHandle);
    FFileHandle:=0;
  end;
  if FReadOnly then begin
    openmode:=GENERIC_READ;
    sharemode:=FILE_SHARE_READ or FILE_SHARE_WRITE;
    createmode:=OPEN_EXISTING;
  end else begin
    openmode:=GENERIC_READ or GENERIC_WRITE;
    sharemode:=FILE_SHARE_READ;
    createmode:=OPEN_ALWAYS;
  end;
  if FileExists(aName) or not(FReadOnly) then begin
    FFileHandle:=agFileOpen(aName,openmode,sharemode,createmode,FILE_ATTRIBUTE_NORMAL);
    if FFileHandle > 0 then begin
      while not agFileEOF(FFileHandle) do
      begin
        agFileReadln(FFileHandle,OneLine);
        if OneLine = #$1A {EOF} then OneLine := '';
        if FFlagDropCommentLines then if OneLine <> '' then if (OneLine[1] = ';') then OneLine := '';
        if OneLine <> '' then begin
          ll := Length(OneLine);
          if (OneLine[1] = '[') and (OneLine[LL] = ']') then
          begin
            CurrSectionName := Copy(OneLine,2,LL-2);
            CurrStringList := TStringList.Create;
            FSectionList.AddObject(CurrSectionName,CurrStringList);
          end
          else begin
            if FFlagDropWhiteSpace then
            begin
              nospace := false;
              repeat
                EqPos := AnsiPos('=', OneLine);
                if EqPos > 1 then begin
                  if OneLine[EqPos - 1] in [' ', #9] then
                    Delete(OneLine, EqPos - 1, 1)
                  else
                    nospace := true;
                end
                else
                  nospace := true;
              until nospace;
              nospace := false;
              EqPos := AnsiPos('=', OneLine);
              if EqPos > 1 then begin
                repeat
                  if EqPos < Length(OneLine) then begin
                    if OneLine[EqPos + 1] in [' ', #9] then
                      Delete(OneLine, EqPos + 1, 1)
                    else
                      nospace := true;
                  end
                  else
                    nospace := true;
                until nospace;
              end;
            end;
            if FFlagDropApostrophes then
            begin
              EqPos := AnsiPos('=', OneLine);
              if EqPos > 1 then begin
                LL := Length(OneLine);
                if EqPos < LL - 1 then begin
                  if (OneLine[EqPos + 1] = OneLine[LL]) and (OneLine[LL] in ['"', #39]) then
                  begin
                    Delete(OneLine, LL, 1);
                    Delete(OneLine, EqPos + 1, 1);
                  end;
                end;
              end;
            end;
            if FFlagTrimRight then
            begin
              lastPos := Length(OneLine);
              while ((lastPos > 0) AND (OneLine[lastPos] < #33)) do Dec(lastPos);
              OneLine := Copy(OneLine,1,lastPos);
            end;
            if (not FFlagFilterOutInvalid) or (AnsiPos('=', OneLine) > 0) then
            begin
              if Assigned(CurrStringList) then CurrStringList.Add(OneLine);
            end;
          end;
        end;
      end;

      if FSectionList.EraseDuplicates(FEraseSectionCallback) then fHasChanged := true;

      if FReadOnly then agFileClose(FFileHandle);
    end else
      Raise EInOutError.CreateFmt('Cannot open file %s',[aName]);
  end;
end;

procedure TagIniFile.SetFileName(const aName: String);
begin
  FlushFile;
  if FFileHandle > 0 then agFileClose(FFileHandle);
  ClearSectionList;
  FFileName:=aName;
  if aName <> '' then AppendFromFile(aName);
  FHasChanged:=False;
end;

function TagIniFile.FindItemIndex(const aSection, aKey: String; CreateNew: Boolean;
  var FoundStringList: TStringList): integer;
var
  SectionIndex: Integer;
  lastIX: Integer;
begin
  SectionIndex := -1;

  if FSectionList.Count > 0 then
  begin
    lastIX := FPrevSectionIndex -1;
    if lastIX < 0 then lastIX := FSectionList.Count -1;
    while (AnsiCompareText(aSection,FSectionList[FPrevSectionIndex]) <> 0)
      and (FPrevSectionIndex <> lastIX) do
    begin
      inc(FPrevSectionIndex);
      if FPrevSectionIndex = FSectionList.Count then FPrevSectionIndex := 0;
    end;
    if AnsiCompareText(aSection,FSectionList[FPrevSectionIndex]) = 0 then
      SectionIndex := FPrevSectionIndex;
  end;

  if SectionIndex = -1 then begin
    if CreateNew then begin
      FoundStringList	:= TStringList.Create;
      FPrevSectionIndex := FSectionList.AddObject(aSection,FoundStringList);
    end
    else
      FoundStringList := NIL;
    Result := -1;
  end else begin
    FoundStringList := FSectionList.SectionItems[SectionIndex];
    Result := FoundStringList.IndexOfName(aKey);
  end;
end;

function TagIniFile.ReadString(const aSection, aKey, aDefault: string): string;
var
  ItemIndex: Integer;
  CurrStringList: TStringList;
begin
  itemIndex := FindItemIndex(aSection,aKey,false,CurrStringList);
  if ItemIndex = -1 then
    result := aDefault
  else
    result := CurrStringList.Values[aKey];
end;

procedure TagIniFile.WriteString(const aSection, aKey, aValue: string);
var
  ItemIndex: Integer;
  CurrStringList: TStringList;
  newLine: String;
begin
  if aKey = '' then
  begin
    // behaviour of WritePrivateProfileString: if all parameters are null strings,
    // the file is flushed to disk. Otherwise, if key name is a null string,
    // the entire Section is to be deleted
    if (aSection = '') and (aValue = '') then FlushFile else EraseSection(aSection);
  end
  else begin
    newLine := aKey+'='+aValue;
    itemIndex := FindItemIndex(aSection,aKey,true,CurrStringList);
    if ItemIndex = -1 then begin
      CurrStringList.Add(newLine);
      FHasChanged   := True;
    end
    else begin
      If (CurrStringList[ItemIndex] <> NewLine) then begin
        FHasChanged := True;
        CurrStringList[ItemIndex] := NewLine;
      end;
    end;
  end;
end;

function TagIniFile.ReadInteger(const aSection, aKey: string;
  aDefault: Longint): Longint;
var
  IStr: string;
begin
  IStr := ReadString(aSection, aKey, '');
  if CompareText(Copy(IStr, 1, 2), '0x') = 0 then
    IStr := '$' + Copy(IStr, 3, 255);
  Result := StrToIntDef(IStr, aDefault);
end;

procedure TagIniFile.WriteInteger(const aSection, aKey: string; aValue: Longint);
begin
  WriteString(aSection, aKey, IntToStr(aValue));
end;

function TagIniFile.ReadBool(const aSection, aKey: string;
  aDefault: Boolean): Boolean;
begin
  Result := ReadInteger(aSection, aKey, Ord(aDefault)) <> 0;
end;

procedure TagIniFile.WriteBool(const aSection, aKey: string; aValue: Boolean);
const
  BoolText: array[Boolean] of string[1] = ('0', '1');
begin
  WriteString(aSection, aKey, BoolText[aValue]);
end;

procedure TagIniFile.ReadSection(const aSection: string; aStrings: TStrings);
var
  SectionIndex,IX: Integer;
  CurrStringList: TStringList;
begin
  SectionIndex := FSectionList.IndexOf(aSection);
  if SectionIndex <> -1 then begin
    CurrStringList := FSectionList.SectionItems[SectionIndex];
    for ix := 0 to CurrStringList.count - 1 do
      aStrings.Add(CurrStringList.Names[IX]);
  end;
end;

procedure TagIniFile.ReadSections(Strings: TStrings);
begin
  strings.assign(SectionNames);
end;

procedure TagIniFile.ReadSectionValues(const aSection: string; aStrings: TStrings);
var
  SectionIndex	  : Integer;
begin
  SectionIndex := FSectionList.IndexOf(aSection);
  if SectionIndex <> -1 then
    aStrings.AddStrings(FSectionList.SectionItems[SectionIndex]);
end;

procedure TagIniFile.ReadAll(aStrings:TStrings);
var
  ixSections: Integer;
  CurrStringList: TStringList;
begin
  with FSectionList do
    for ixSections := 0 to count -1 do begin
      CurrStringList := SectionItems[ixSections];
      if CurrStringList.Count > 0 then begin
	aStrings.Add('['+Strings[ixSections]+']');
	aStrings.AddStrings(CurrStringList);
	aStrings.Add('');
      end;
    end;
end;

procedure TagIniFile.FlushFile;
var
  CurrStringList: TStringList;
  ix,ixSections: Integer;
begin
  if FHasChanged then begin
    if FFileName <> '' then
      if (FFileHandle > 0) and not(FReadOnly) then begin
        FileSeek(FFileHandle,0,0);
        with FSectionList do
          for ixSections := 0 to count -1 do begin
            CurrStringList := SectionItems[ixSections];
            if CurrStringList.Count > 0 then begin
              agFileWriteln(FFileHandle,'['+Strings[ixSections]+']');
              for ix := 0 to CurrStringList.Count-1 do
                agFileWriteln(FFileHandle,CurrStringList[ix]);
              agFileWriteln(FFileHandle,'');
            end;
          end;
        SetEndOfFile(FFileHandle);
      end;
    FHasChanged   := False;
  end;
end;

procedure TagIniFile.EraseSection(const aSection: string);
var
  SectionIndex: Integer;
begin
  SectionIndex := FSectionList.IndexOf(aSection);
  if SectionIndex <> -1 then begin
    FSectionList.SectionItems[SectionIndex].Free;
    FSectionList.Delete(SectionIndex);
    FSectionList.FPrevIndex := 0;
    FHasChanged := True;
  end;
end;

procedure TagIniFile.DeleteKey(const aSection, aKey: string);
var
  ItemIndex: Integer;
  CurrStringList: TStringList;
begin
  itemIndex := FindItemIndex(aSection,aKey,true,CurrStringList);
  if ItemIndex > -1 then begin
    FHasChanged := True;
    CurrStringList.Delete(ItemIndex);
  end;
end;

end.