{*******************************************************}
{                                                       }
{       Alex Ghost Library                              }
{                                                       }
{       Copyright (c) 1999,2000 Alexey Popov            }
{                                                       }
{*******************************************************}

{$I AG.INC}

unit agTranslator;

interface

uses Windows, Classes, Forms, agIniFile;

type
  TSetLanguageEvent = procedure(Sender: TObject; const Lang: string;
    LangIndex: integer) of object;

  TLangResOffsetEvent = procedure(Sender: TObject; const Lang: string;
    LangIndex: integer; var LangResOffset: integer) of object;

  TUpdateCompLangEvent = procedure(Sender: TObject; Component: TComponent;
    var LangStr: string; var Handled: boolean) of object;

  TagTranslator = class(TComponent)
  private
    FDir, FFullDir, FLang: string;
    FLangIndex, FResOffs: integer;
    FLangFileList,FLangList: TStringList;
    FLngFile: TagIniFile;
    FUseTag: boolean;
    FOnSetLanguage: TSetLanguageEvent;
    FOnLangResOffsetEvent: TLangResOffsetEvent;
    FUpdateCompLangEvent: TUpdateCompLangEvent;
    procedure SetDir(const Dir: string);
    procedure SetLang(const Lang: string);
    procedure SetLangIndex(Index: integer);
    procedure LoadLangList;
  protected
    procedure Loaded; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    property LanguageList: TStringList read FLangList;
    property LangName: string read FLang write SetLang;
    property LangIndex: integer read FLangIndex write SetLangIndex;
    procedure UpdateFormLanguage(Form: TForm);
    procedure UpdateComponentLanguage(Component: TComponent);
    function LoadLangStr(const StrID: string): string;
    function LoadLangStrSect(const Section,StrID: string): string;
    function LoadLangStrID(ID: integer): string;
  published
    property Directory: string read FDir write SetDir;
    property UseTag: boolean read FUseTag write FUseTag default true;
    property OnSetLanguage: TSetLanguageEvent read FOnSetLanguage write FOnSetLanguage;
    property OnLangResOffset: TLangResOffsetEvent read FOnLangResOffsetEvent
      write FOnLangResOffsetEvent;
    property OnUpdateComponentLanguage: TUpdateCompLangEvent read FUpdateCompLangEvent
      write FUpdateCompLangEvent;
  end;

var
  MainTranslator: TagTranslator;

procedure SetLangIndex(Index: integer);
procedure SetLangName(const Lang: string);
procedure UpdateFormLanguage(Form: TForm);
procedure UpdateComponentLanguage(Component: TComponent);
function LoadLangStr(const StrID: string): string;
function LoadLangStrSect(const Section,StrID: string): string;
function LoadLangStrID(ID: integer): string;

implementation

uses SysUtils, Messages, StdCtrls, ComCtrls, Buttons, Menus, ExtCtrls, Controls,
  {$IFDEF RX} ToolEdit, {$ENDIF} agFileFind, agFileUtils, agStrUtils, agPropUtils;

const
  CStringsSection        = 'Strings';
  CLanguageName          = 'LanguageName';
  CCaptHintDiv           = '~';
  CItemsDiv              = '|';
  CLineDiv               = '||';
  NeedSetLang: integer   = 0;
  NeedSetLangStr: string = '';
  FNeedSetLang: boolean  = false;
  FWndHook: HHook        = 0;

type
  THackListView = class(TCustomListView);
  THackTabControl = class(TCustomTabControl);

function GetLngCaption(const s: string): string;
begin
  Result:=LeftPart(s,CCaptHintDiv);
end;

function GetLngHint(const s: string): string;
begin
  Result:=RightPart(s,CCaptHintDiv);
end;

function GetLngItem(const s: string; i: integer): string;
begin
  Result:=GetWord(i,s,[CItemsDiv]);
end;

procedure TagTranslator.LoadLangList;
var
  i: integer;
  sr: TSearchRecEx;
  s: string;
  Lng: TagIniFile;
begin
  FLangFileList.Clear;
  FLangList.Clear;
  i:=FindFirstEx(Format('%s*.lng',[FFullDir]),0,faNotFiles,sr);
  try
    while i = 0 do begin
      Lng:=TagIniFile.Create(Format('%s%s',[FFullDir,sr.Name]),true);
      try
        s:=Lng.ReadString(CStringsSection,CLanguageName,'');
        if s <> '' then begin
          FLangFileList.Add(Format('%s=%s',[s,sr.Name]));
          FLangList.Add(s);
        end;
      finally
        Lng.Free;
      end;
      i:=FindNextEx(sr);
    end;
  finally
    FindCloseEx(sr);
  end;
end;

function ShowWindowHook(Code: Integer; WParam: WPARAM; LParam: LPARAM): LRESULT; stdcall;
var
  C: TWinControl;
begin
  if Code = HC_ACTION then begin
    if (PCWPStruct(LParam)^.message = WM_SHOWWINDOW) and
      (PCWPStruct(LParam)^.wParam <> 0) then
    begin
      C:=FindControl(PCWPStruct(LParam)^.hwnd);
      if Assigned(C) then
        if C is TForm then UpdateFormLanguage(C as TForm);
    end;
    Result:=CallNextHookEx(FWndHook,Code,WParam,LParam);
  end else
    Result:=0;
end;

constructor TagTranslator.Create(AOwner: TComponent);
begin
  if Assigned(MainTranslator) then
    raise EInvalidOperation.Create('Only one instance of TagLanguage is allowed');
  inherited Create(AOwner);
  FUseTag:=true;
  Directory:='';
  FLangFileList:=TStringList.Create;
  FLangList:=TStringList.Create;
  FLangList.Sorted:=true;
  if MainTranslator = nil then MainTranslator:=Self;
  FWndHook:=SetWindowsHookEx(WH_CALLWNDPROC,ShowWindowHook,0,GetCurrentThreadId);
end;

destructor TagTranslator.Destroy;
begin
  UnhookWindowsHookEx(FWndHook);
  if MainTranslator = Self then MainTranslator:=nil;
  if Assigned(FLngFile) then FLngFile.Free;
  FLangList.Free;
  FLangFileList.Free;
  inherited Destroy;
end;

function TagTranslator.LoadLangStrSect(const Section,StrID: string): string;
begin
  if Assigned(FLngFile) then begin
    Result:=ReplaceStr(FLngFile.ReadString(Section,StrID,''),CLineDiv,#13#10);
  end else
    Result:='';
end;

function TagTranslator.LoadLangStr(const StrID: string): string;
begin
  Result:=LoadLangStrSect(CStringsSection,StrID);
end;

function TagTranslator.LoadLangStrID(ID: integer): string;
begin
  Result:=LoadLangStrSect(CStringsSection,IntToStr(ID));
  if Result = '' then begin
    Result:=ReplaceStr(LoadStr(FResOffs+ID),CLineDiv,#13#10);
  end;
end;

procedure TagTranslator.SetLang(const Lang: string);
var
  i: integer;
  s,fn: string;
begin
  if csDesigning in ComponentState then exit;
  if FLangList.Count = 0 then exit;

  fn:=FLangFileList.Values[Lang];
  if fn = '' then exit;

  FLang:=Lang;
  FLangIndex:=FLangList.IndexOf(Lang);
  FResOffs:=0;
  if Assigned(FOnLangResOffsetEvent) then
    FOnLangResOffsetEvent(Self,FLang,FLangIndex,FResOffs);

  if Assigned(FLngFile) then begin
    FLngFile.Free;
    FLngFile:=nil;
  end;
  s:=Format('%s%s',[FFullDir,fn]);
  if FileExists(s) then FLngFile:=TagIniFile.Create(s,true);

  for i:=0 to Screen.FormCount-1 do UpdateFormLanguage(Screen.Forms[i]);

  if Assigned(FOnSetLanguage) then FOnSetLanguage(Self,FLang,FLangIndex);
end;

procedure TagTranslator.SetLangIndex(Index: integer);
begin
  if (Index >= 0) and (Index < FLangList.Count) then begin
    FLangIndex:=Index;
    SetLang(FLangList[Index]);
  end;
end;

procedure TagTranslator.UpdateComponentLanguage(Component: TComponent);
var
  j: integer;
  s,s1,sOwner: string;
  C: TComponent;
  Handled: boolean;
begin
  if csDesigning in ComponentState then exit;
  C:=Component;

  if Assigned(C.Owner) then sOwner:=C.Owner.ClassName else sOwner:=CStringsSection;
  s:=LoadLangStrSect(sOwner,C.Name);
  if (s = '') and (C.Tag > 0) and FUseTag then s:=LoadLangStrID(C.Tag);

  Handled:=false;
  if Assigned(FUpdateCompLangEvent) then FUpdateCompLangEvent(Self,C,s,Handled);
  if Handled then exit;

  if s <> '' then begin
    // specific processing
    // TPageControl
    if C is TPageControl then begin
      for j:=0 to (C as TPageControl).PageCount-1 do begin
        s1:=GetLngItem(s,j+1);
        SetStrProperty((C as TPageControl).Pages[j],'Caption',GetLngCaption(s1));
        SetStrProperty((C as TPageControl).Pages[j],'Hint',GetLngHint(s1));
      end;
      exit;
    end;
    // TListView
    if C is TCustomListView then begin
      for j:=0 to THackListView(C).Columns.Count-1 do begin
        s1:=GetLngItem(s,j+1);
        SetStrProperty(THackListView(C).Columns[j],'Caption',GetLngCaption(s1));
      end;
      exit;
    end;
    // TTabControl
    if C is TCustomTabControl then begin
      for j:=0 to THackTabControl(C).Tabs.Count-1 do begin
        s1:=GetLngItem(s,j+1);
        THackTabControl(C).Tabs[j]:=GetLngCaption(s1);
      end;
      exit;
    end;

    {$IFDEF RX}
    // RXLib classes
    // TDirectoryEdit
    if C is TDirectoryEdit then begin
      (C as TDirectoryEdit).DialogText:=GetLngCaption(s);
      (C as TDirectoryEdit).ButtonHint:=GetLngHint(s);
      exit;
    end;
    {$ENDIF}

    // common processing
    SetStrProperty(C,'Caption',GetLngCaption(s));
    SetStrProperty(C,'Hint',GetLngHint(s));
  end;
end;

procedure TagTranslator.UpdateFormLanguage(Form: TForm);
var
  i: integer;
  s: string;
  Handled: boolean;
begin
  if csDesigning in ComponentState then exit;

  // form
  s:=LoadLangStrSect(Form.ClassName,'Self');
  if (s = '') and (Form.Tag > 0) and FUseTag then s:=LoadLangStrID(Form.Tag);

  Handled:=false;
  if Assigned(FUpdateCompLangEvent) then FUpdateCompLangEvent(Self,Form,s,Handled);
  if Handled then exit;

  if s <> '' then begin
    Form.Caption:=GetLngCaption(s);
    Form.Hint:=GetLngHint(s);
  end;

  // form components
  for i:=0 to Form.ComponentCount-1 do
    UpdateComponentLanguage(Form.Components[i]);
end;

procedure TagTranslator.SetDir(const Dir: string);
begin
  FDir:=Dir;
  FFullDir:=NormalDir(ExpandRelativePath(Dir,ExtractFilePath(ParamStr(0))));
end;

procedure TagTranslator.Loaded;
begin
  LoadLangList;
  if FNeedSetLang then begin
    if NeedSetLangStr = '' then
      LangIndex:=NeedSetLang
    else
      LangName:=NeedSetLangStr;
  end else
    if FLangList.Count > 0 then LangIndex:=0;
end;

procedure SetLangName(const Lang: string);
begin
  if Assigned(MainTranslator) then
    MainTranslator.LangName:=Lang
  else begin
    NeedSetLangStr:=Lang;
    FNeedSetLang:=true;
  end;
end;

procedure SetLangIndex(Index: integer);
begin
  if Assigned(MainTranslator) then
    MainTranslator.LangIndex:=Index
  else begin
    NeedSetLang:=Index;
    FNeedSetLang:=true;
  end;
end;

procedure UpdateFormLanguage(Form: TForm);
begin
  if Assigned(MainTranslator) then
    MainTranslator.UpdateFormLanguage(Form);
end;

procedure UpdateComponentLanguage(Component: TComponent);
begin
  if Assigned(MainTranslator) then
    MainTranslator.UpdateComponentLanguage(Component);
end;

function LoadLangStr(const StrID: string): string;
begin
  if Assigned(MainTranslator) then
    Result:=MainTranslator.LoadLangStr(StrID);
end;

function LoadLangStrSect(const Section,StrID: string): string;
begin
  if Assigned(MainTranslator) then
    Result:=MainTranslator.LoadLangStrSect(Section,StrID);
end;

function LoadLangStrID(ID: integer): string;
begin
  if Assigned(MainTranslator) then
    Result:=MainTranslator.LoadLangStrID(ID);
end;

end.
