UNIT dropsource;
  
  // -----------------------------------------------------------------------------
  // Project:         Drag and Drop Source Components
  // Component Names: TDropTextSource, TDropFileSource
  // Module:          DropSource
  // Description:     Implements Dragging & Dropping of text and files
  //                  FROM your application to another.
  // Version:	        3.0
  // Date:            22-SEP-1998
  // Target:          Win32, Delphi 3 & 4
  // Authors:         Angus Johnson, ajohnson@rpi.net.au
  //                  Anders Melander, anders@melander.dk
  //                                   http://www.melander.dk
  // Copyright        1998 Angus Johnson & Anders Melander
  // -----------------------------------------------------------------------------
  // You are free to use this source but please give us credit for our work.
  // if you make improvements or derive new components from this code,
  // we would very much like to see your improvements. FEEDBACK IS WELCOME.
  // -----------------------------------------------------------------------------

  // History:
  // dd/mm/yy  Version  Changes
  // --------  -------  ----------------------------------------
  // 22.09.98  3.0      * Shortcuts (links) for TDropFileSource now enabled.
  //                    * Scrap files for TDropTextSource now enabled.
  //                    * DropSource property added. If DropSource is assigned then
  //                      its MouseDown and MouseMove logic is overridden
  //                      to initiate an auto drag op.
  //                    * New events - OnStartDrag and OnEndDrag added (See demo).
  //                      OnStartDrag is called automatically if DropSource
  //                      Control is assigned otherwise use Execute function to
  //                      manually initiate a drag op.
  //                    * TDropSource.DoEnumFormatEtc() no longer declared abstract.
  // 08.09.98  2.0      * No significant changes to this module
  //                      but the version was updated to coincide with the
  //                      new DropTarget module included with this demo.
  // 31.08.98  1.5      * Fixed a Delphi 4 bug!
  //                      (I cut and pasted the wrong line!)
  //                    * Demo code now MUCH tidier and easier to read (I think).
  // 19.08.98  1.4      * CopyToClipboard method added.
  //                    * Should now compile in Delphi 4. (see below)
  //                    * Another tidy up of the code.
  // 21.07.98  1.3      * Fixed a bug in Ver 1.2 where OnDrop event was never called.
  //                    * Now able to drag text to WordPad.
  //                    * Added OnFeedback Event.
  //                    * Added dtLink to TDragType enumeration but
  //                      still not able to get it to work.
  //                    * Code tidy up.
  // 19.07.98  1.2      * Changed TDragType enumeration type to a
  //                      TDragTypes set (AM's suggestion) ready to
  //                      implement dtLink later.
  //                    * Added known bugs to header and demo.
  // 17.07.98  1.1      * Reenabled end-user option to select either
  //                      Copy or Move operation while dragging.
  // 15.07.98  1.0      * Initial Delphi 3 component implementation of AM's
  //                      DropSource unit. I released a Delphi 2 D'n'D
  //                      component (TDragFilesSrc) over 12 months ago. However
  //                      with the significant changes in COM implementations
  //                      between Delphi versions I decided to use AM's code
  //                      as the springboard for my new Delphi 3 D'n'D components.
  //                      Thanks to Anders for the excellent start and
  //                      suggestions along the way!
  //
  // -----------------------------------------------------------------------------

  // TDropTextSource -
  //   Public
  //      ....
  //      Text: string;
  //      function Execute: TDragResult; //drDropCopy, drDropMove, drDropCancel ...
  //      function CopyToClipboard: boolean;
  //    published
  //      property DragTypes: TDragTypes;  // [dtCopy, dtMove, dtLink]
  //      property OnDrop: TDropEvent;
  //      property OnFeedback: TFeedbackEvent;
  //      ....

  // TDropFileSource -
  //   Public
  //      ....
  //      property Files: TStrings;
  //      function Execute: TDragResult; //drDropCopy, drDropMove, drDropCancel ...
  //      function CopyToClipboard: boolean;
  //    published
  //      property DragTypes: TDragTypes;  // [dtCopy, dtMove]
  //      property OnDrop: TDropEvent;
  //      property OnFeedback: TFeedbackEvent;
  //      ....

  //
  // Very brief examples of usage (see included demo for more detailed examples):
  //
  // TDropTextSource -
  // DropFileSource1.DragTypes := [dtCopy];
  // DropTextSource1.text := edit1.text;
  // if DropTextSource1.execute = drDropCopy then ShowMessage('It worked!');
  //
  // TDropFileSource -
  // DropFileSource1.DragTypes := [dtCopy, dtMove];
  // //  ie: let the user decide - Copy or Move.
  // //      Hold Ctrl down during drag -> Copy
  // //      Hold Shift down during drag -> Move
  // DropFileSource1.files.clear;
  // DropFileSource1.files.add('c:\autoexec.bat');
  // DropFileSource1.files.add('c:\config.sys');
  // res := DropFileSource1.execute;
  // if res = drDropCopy then ShowMessage('Files Copied!')
  // else if res = drDropMove then ShowMessage('Files Moved!');
  // -----------------------------------------------------------------------------

INTERFACE
  USES
    Controls, Windows, ActiveX, Classes, ShlObj, SysUtils, ClipBrd;

  const
    MaxFormats = 20;

  TYPE

  TInterfacedComponent = class(TComponent, IUnknown)
  Private
    fRefCount: Integer;
  Protected
    function QueryInterface(const IID: TGuid; OUT Obj): HRESULT;
               {$ifdef VER120} reintroduce; {$endif} StdCall;
    function _AddRef: Integer; StdCall;
    function _Release: Integer; StdCall;
  Public
     property RefCount: Integer Read fRefCount;
  end;

  TDragType = (dtCopy, dtMove, dtLink);
  TDragTypes = SET OF TDragType;

  TDragResult = (drDropCopy, drDropMove, drDropLink, drCancel, drOutMemory, drUnknown);

  TDropEvent = procedure(Sender: TObject; DragType: TDragType;
               var ContinueDrop: Boolean) OF Object;
  TFeedbackEvent = procedure(Sender: TObject; Effect: LongInt) OF Object;

  TStartDragEvent = procedure (Sender: TObject; var Continue: boolean) of object; 
  TEndDragEvent = procedure (Sender: TObject; Result: TDragResult) of object; 

  TDropSource = class(TInterfacedComponent, IDropSource, IDataObject)
  Private
    fDragTypes      : TDragTypes;
    FeedbackEffect  : LongInt;
    fDropEvent      : TDropEvent;
    fFBEvent        : TFeedBackEvent;
    fDataFormats    : array[0..MaxFormats-1] of TFormatEtc;
    DataFormatsCount: integer;
    FDropSource	 : TControl; 
    FOnMouseDown	 : TMouseEvent; 
    FOnMouseMove	 : TMouseMoveEvent; 
    FOnStartDrag	 : TStartDragEvent; 
    FOnEndDrag		 : TEndDragEvent; 
    FDragPoint		 : TPoint; 
  Protected
    // IDropSource implementation

    function QueryContinueDrag(fEscapePressed: bool; grfKeyState: LongInt): HRESULT; StdCall;
    function GiveFeedback(dwEffect: LongInt): HRESULT; StdCall;

    // IDataObject implementation
    function GetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT; StdCall;
    function GetDataHere(const FormatEtc: TFormatEtc; OUT Medium: TStgMedium):HRESULT; StdCall;
    function QueryGetData(const FormatEtc: TFormatEtc): HRESULT; StdCall;
    function GetCanonicalFormatEtc(const FormatEtc: TFormatEtc;
             OUT FormatEtcOut: TFormatEtc): HRESULT; StdCall;
    function SetData(const FormatEtc: TFormatEtc; var Medium: TStgMedium;
             fRelease: Bool): HRESULT; StdCall;
    function EnumFormatEtc(dwDirection: LongInt; OUT EnumFormatEtc: IEnumFormatEtc): HRESULT; StdCall;
    function dAdvise(const FormatEtc: TFormatEtc; advf: LongInt;
             const advsink: IAdviseSink; OUT dwConnection: LongInt): HRESULT; StdCall;
    function dUnadvise(dwConnection: LongInt): HRESULT; StdCall;
    function EnumdAdvise(OUT EnumAdvise: IEnumStatData): HRESULT; StdCall;

    //New functions...
    function DoGetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Virtual;
    function DoGetDataHere(const FormatEtc: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Virtual;
    function DoQueryGetData(const FormatEtc: TFormatEtc): HRESULT; Virtual;
    function DoEnumFormatEtc(dwDirection: LongInt; OUT EnumFormatEtc: IEnumFormatEtc): HRESULT; Virtual;

    procedure SetDropSource(const Value: TControl); 
    procedure Loaded; override; 
    procedure Notification(AComponent: TComponent; Operation: TOperation); override; 
    procedure Hook; 
    procedure Unhook; 
    procedure DoOnMouseDown(Sender: TObject; Button: TMouseButton;
      Shift: TShiftState; X, Y: Integer); 
    procedure DoOnMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer); 
  Public
    constructor Create(aowner: TComponent); Override;
    destructor Destroy; override;
    function Execute: TDragResult;
    function CopyToClipboard: boolean; Virtual;
  Published
    property Dragtypes: TDragTypes read fDragTypes write fDragTypes;
    property DropSource: TControl read FDropSource write SetDropSource; 
    property OnStartDrag: TStartDragEvent read FOnStartDrag write FOnStartDrag; 
    property OnFeedback: TFeedbackEvent Read fFBEvent Write fFBEvent;
    property OnDrop: TDropEvent Read fDropEvent Write fDropEvent;
    property OnEndDrag: TEndDragEvent read FOnEndDrag write FOnEndDrag; 
  end;

  TDropTextSource = class(TDropSource)
  Private
    fText: String;
  Protected
    function DoGetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Override;
    function DoGetDataHere(const FormatEtc: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Override;
  Public
    constructor Create(aOwner: TComponent); Override;
    function CopyToClipboard: boolean; Override;
    property Text: String Read fText Write fText;
  end;

  TDropFileSource = class(TDropSource)
  Private
    fFiles: TStrings;
  Protected
    function DoGetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Override;
    function DoGetDataHere(const FormatEtc: TFormatEtc; OUT Medium: TStgMedium):HRESULT; Override;
  Public
    constructor Create(aOwner: TComponent); Override;
    destructor Destroy; Override;
    function CopyToClipboard: boolean; Override;
    property Files: TStrings Read fFiles;
  end;

//  procedure Register;

IMPLEMENTATION

  // -----------------------------------------------------------------------------
  //			Miscellaneous declarations and functions.
  // -----------------------------------------------------------------------------

  TYPE
    TMyDropFiles = PACKED RECORD
      dropfiles: TDropFiles;
      filenames: ARRAY[1..1] OF Char;
    end;
    pMyDropFiles = ^TMyDropFiles;


  //******************* GetSizeOfPidl *************************
  function GetSizeOfPidl(pidl: pItemIDList): integer;
  var
    ptr: PByte;
    pSHIt: pSHItemID;
    i: integer;
  begin
    result := 2;
    ptr := pointer(pidl);
    repeat
      pSHIt := pointer(ptr);
      i := pSHIt^.cb;
      inc(result,i);
      inc(ptr,i);
    until i = 0;
  end;

  //******************* FreePidl *************************
  procedure FreePidl(pidl: pItemIDList);
  var
    ShellMalloc: IMalloc;
  begin
    SHGetMalloc(ShellMalloc);
    ShellMalloc.free(pidl);
  end;

  //******************* GetShellFolderOfPath *************************
  function GetShellFolderOfPath(FolderPath: TFileName): IShellFolder;
  var
    DeskTopFolder: IShellFolder;
    PathPidl: pItemIDList;
    OlePath: Array[0..MAX_PATH] of WideChar;
    dummy,pdwAttributes: ULONG;
  begin
    result := nil;
    StringToWideChar( FolderPath, OlePath, MAX_PATH );
    try
      if not (SHGetDesktopFolder(DeskTopFolder) = NOERROR) then exit;
      if (DesktopFolder.ParseDisplayName(0,
            nil,OlePath,dummy,PathPidl,pdwAttributes) = NOERROR) and
            (pdwAttributes and SFGAO_FOLDER <> 0) then
        DesktopFolder.BindToObject(PathPidl,nil,IID_IShellFolder,pointer(result));
      FreePidl(PathPidl);
    except
    end;
  end;

  //******************* GetFullPIDLFromPath *************************
  function GetFullPIDLFromPath(Path: TFileName): pItemIDList;
  var
     DeskTopFolder: IShellFolder;
     OlePath: Array[0..MAX_PATH] of WideChar;
     dummy1,dummy2: ULONG;
  begin
    result := nil;
    StringToWideChar( Path, OlePath, MAX_PATH );
    try
      if (SHGetDesktopFolder(DeskTopFolder) = NOERROR) then
        DesktopFolder.ParseDisplayName(0,nil,OlePath,dummy1,result,dummy2);
    except
    end;
  end;

  //******************* GetSubPidl *************************
  function GetSubPidl(Folder: IShellFolder; Sub: TFilename): pItemIDList;
  var
    dummy1,dummy2: ULONG;
    OleFile: Array[0..MAX_PATH] of WideChar;
  begin
    result := nil;
    try
      StringToWideChar( Sub, OleFile, MAX_PATH );
      Folder.ParseDisplayName(0,nil,OleFile,dummy1,result,dummy2);
    except
    end;
  end;

  //See "Clipboard Formats for Shell Data Transfers" in Ole.hlp...
  //(Needed to drag links (shortcuts).)

  //******************* ConvertFilesToShellIDList *************************
  type
    POffsets = ^TOffsets;
    TOffsets = array[0..$FFFF] of UINT;

  function ConvertFilesToShellIDList(path: string; files: TStrings): HGlobal;
  var
    shf: IShellFolder;
    PathPidl, pidl: pItemIDList;
    Ida: PIDA;
    pOffset: POffsets;
    ptrByte: ^Byte;
    i, PathPidlSize, IdaSize, PreviousPidlSize: integer;
  begin
    result := 0;
    shf := GetShellFolderOfPath(path);
    if shf = nil then exit;
    //Calculate size of IDA structure ...
    IdaSize := (files.count + 2) * sizeof(UINT);

    //Add to IdaSize space for ALL pidls...
    PathPidl := GetFullPIDLFromPath(path);
    PathPidlSize := GetSizeOfPidl(PathPidl);
    IdaSize := IdaSize + PathPidlSize;
    for i := 0 to files.count-1 do
    begin
      pidl := GetSubPidl(shf,files[i]);
      IdaSize := IdaSize + GetSizeOfPidl(Pidl);
      FreePidl(pidl);
    end;

    //Allocate memory...
    Result := GlobalAlloc(GMEM_SHARE or GMEM_ZEROINIT, IdaSize);
    if (Result = 0) then
    begin
      FreePidl(PathPidl);
      Exit;
    end;
    Ida := GlobalLock(Result);
    FillChar(Ida^,IdaSize,0);

    //Fill in offset and pidl data...
    Ida^.cidl := files.count; //cidl = file count
    pOffset := @(Ida^.aoffset); //otherwise I would have to turn off range checking.
    pOffset^[0] := (files.count+2)*sizeof(UINT); //offset of Path pidl

    ptrByte := pointer(Ida);
    inc(ptrByte,pOffset^[0]); //ptrByte now points to Path pidl
    move(PathPidl^, ptrByte^, PathPidlSize); //copy path pidl
    FreePidl(PathPidl);

    PreviousPidlSize := PathPidlSize;
    for i := 1 to files.count do
    begin
      pidl := GetSubPidl(shf,files[i-1]);
      pOffset^[i] := pOffset^[i-1] + UINT(PreviousPidlSize); //offset of pidl
      PreviousPidlSize := GetSizeOfPidl(Pidl);

      ptrByte := pointer(Ida);
      inc(ptrByte,pOffset^[i]); //ptrByte now points to current file pidl
      move(Pidl^, ptrByte^, PreviousPidlSize); //copy file pidl
                            //PreviousPidlSize = current pidl size here
      FreePidl(pidl);
    end;
    GlobalUnLock(Result);
  end;


  //******************* Register *************************
  procedure Register;
  begin
    RegisterComponents('Samples', [TDropFileSource,TDropTextSource]);
  end;


  // -----------------------------------------------------------------------------
  //			TInterfacedComponent
  // -----------------------------------------------------------------------------

  // QueryInterface now returns HRESULT (needed for Delphi 4).
  // Thanks to 'Scotto the Unwise' - scottos@gtcom.net
  //******************* TInterfacedComponent.QueryInterface *************************
  function TInterfacedComponent.QueryInterface(const IID: TGuid; OUT Obj): HRESULT;
  begin
    if GetInterface(IID, Obj) then result := 0 else result := E_NOINTERFACE;
  end;

  //******************* TInterfacedComponent._AddRef *************************
  function TInterfacedComponent._AddRef: Integer;
  begin
    Inc(fRefCount);
    result := fRefCount;
  end;

  //******************* TInterfacedComponent._Release *************************
  function TInterfacedComponent._Release: Integer;
  begin
    Dec(fRefCount);
    if fRefCount = 0 then
    begin
      Destroy;
      result := 0;
      Exit;
    end;
    result := fRefCount;
  end;

  // -----------------------------------------------------------------------------
  //			TEnumFormatEtc
  // -----------------------------------------------------------------------------
  { TEnumFormatEtc - format enumerator for TDataObject }

  TYPE
    pFormatList = ^TFormatList;
    TFormatList = ARRAY[0..255] OF TFormatEtc;

  TEnumFormatEtc = class(TInterfacedObject, IEnumFormatEtc)
  Private
    fFormatList: pFormatList;
    fFormatCount: Integer;
    fIndex: Integer;
  Public
    constructor Create(FormatList: pFormatList; FormatCount, Index: Integer);
    { IEnumFormatEtc }
    function Next(Celt: LongInt; OUT Elt; pCeltFetched: pLongInt): HRESULT; StdCall;
    function Skip(Celt: LongInt): HRESULT; StdCall;
    function Reset: HRESULT; StdCall;
    function Clone(OUT Enum: IEnumFormatEtc): HRESULT; StdCall;
  end;

//******************* TEnumFormatEtc.Create *************************
  constructor TEnumFormatEtc.Create(FormatList: pFormatList;
              FormatCount, Index: Integer);
  begin
    inherited Create;
    fFormatList := FormatList;
    fFormatCount := FormatCount;
    fIndex := Index;
  end;

  //******************* TEnumFormatEtc.Next *************************
  function TEnumFormatEtc.Next(Celt: LongInt; OUT Elt; pCeltFetched: pLongInt): HRESULT;
  var
    i: Integer;
  begin
    i := 0;
    WHILE (i < Celt) and (fIndex < fFormatCount) DO
    begin
      TFormatList(Elt)[i] := fFormatList[fIndex];
      Inc(fIndex);
      Inc(i);
    end;
    if pCeltFetched <> NIL then pCeltFetched^ := i;
    if i = Celt then result := S_OK else result := S_FALSE;
  end;

  //******************* TEnumFormatEtc.Skip *************************
  function TEnumFormatEtc.Skip(Celt: LongInt): HRESULT;
  begin
    if Celt <= fFormatCount - fIndex then
    begin
      fIndex := fIndex + Celt;
      result := S_OK;
    end else
    begin
      fIndex := fFormatCount;
      result := S_FALSE;
    end;
  end;

  //******************* TEnumFormatEtc.Reset *************************
  function TEnumFormatEtc.ReSet: HRESULT;
  begin
    fIndex := 0;
    result := S_OK;
  end;

  //******************* TEnumFormatEtc.Clone *************************
  function TEnumFormatEtc.Clone(OUT Enum: IEnumFormatEtc): HRESULT;
  begin
    enum := TEnumFormatEtc.Create(fFormatList, fFormatCount, fIndex);
    result := S_OK;
  end;

  // -----------------------------------------------------------------------------
  //			TDropSource
  // -----------------------------------------------------------------------------

  //******************* TDropSource.Create *************************
  constructor TDropSource.Create(aOwner: TComponent);
  begin
    inherited Create(aOwner);
    DragTypes := [dtCopy]; //default to Copy.
    //to avoid premature release ...
    _AddRef;
    DataFormatsCount := 0;
  end;

  //******************* TDropSource.Destroy *************************
  destructor TDropSource.Destroy; 
  begin
    Unhook;
    inherited Destroy;
  end;

  //******************* TDropSource.Execute *************************
  function TDropSource.Execute: TDragResult;
  var
    res: HRESULT;
    okeffect: LongInt;
    effect: LongInt;
  begin
    result := drUnknown;
    okeffect := DROPEFFECT_NONE;
    if dtCopy IN fDragTypes then okeffect := okeffect + DROPEFFECT_COPY;
    if dtMove IN fDragTypes then okeffect := okeffect + DROPEFFECT_MOVE;
    if dtLink IN fDragTypes then okeffect := okeffect + DROPEFFECT_LINK;

    res := DoDragDrop(Self AS IDataObject, Self AS IDropSource, okeffect, effect);
    CASE res OF
      DRAGDROP_S_DROP:   begin
                           if (okeffect and effect <> 0) then
                           begin
                             if (effect and DROPEFFECT_COPY <> 0) then
                               result := drDropCopy
                             else if (effect and DROPEFFECT_MOVE <> 0) then
                               result := drDropMove
                             else result := drDropLink;
                           end else
                             result := drCancel;
                         end;
      DRAGDROP_S_CANCEL: result := drCancel;
      E_OUTOFMEMORY:     result := drOutMemory;
    end;
    if (Assigned(FOnEndDrag)) then 
      FOnEndDrag(Self, Result);
  end;

  //******************* TDropSource.CopyToClipboard *************************
  function TDropSource.CopyToClipboard: boolean;
  begin
    result := false;
  end;

  //******************* TDropSource.QueryContinueDrag *************************
  function TDropSource.QueryContinueDrag(fEscapePressed: bool;
    grfKeyState: LongInt): HRESULT; StdCall;
  var
    ContinueDrop: Boolean;
    dragtype: TDragType;
  begin
    if fEscapePressed then
      result := DRAGDROP_S_CANCEL
    // will now allow drag and drop with either mouse button.
    else if (grfKeyState and (MK_LBUTTON or MK_RBUTTON) = 0) then
    begin
      ContinueDrop := True;
      dragtype := dtCopy;
      if (FeedbackEffect and DROPEFFECT_COPY <> 0) then DragType := dtCopy
      else if (FeedbackEffect and DROPEFFECT_MOVE <> 0) then dragtype := dtMove
      else if (FeedbackEffect and DROPEFFECT_LINK <> 0) then dragtype := dtLink
      else ContinueDrop := False;

      //if a valid drop then do OnDrop event if assigned...
      if ContinueDrop and
        (DragType in [dtCopy, dtMove, dtLink] * dragtypes) and 
        Assigned(OnDrop) then OnDrop(Self, dragtype, ContinueDrop);

      if ContinueDrop then result := DRAGDROP_S_DROP
      else result := DRAGDROP_S_CANCEL;
    end else
      result := NOERROR;
  end;

  //******************* TDropSource.GiveFeedback *************************
  function TDropSource.GiveFeedback(dwEffect: LongInt): HRESULT; StdCall;
  begin
    FeedbackEffect := dwEffect;
    //NB: Use the OnFeedback event sparingly as it will effect performance...
    if Assigned(OnFeedback) then OnFeedback(Self, dwEffect);
    result:=DRAGDROP_S_USEDEFAULTCURSORS;
  end;

  //******************* TDropSource.DoQueryGetData *************************
  function TDropSource.DoQueryGetData(const FormatEtc: TFormatEtc): HRESULT;
  var
    i: integer;
  begin
    result:= S_OK;
    for i := 0 to DataFormatsCount-1 do
      with fDataFormats[i] do
      begin
        if (FormatEtc.cfFormat = cfFormat) and
           (FormatEtc.dwAspect = dwAspect) and
           (FormatEtc.tymed and tymed <> 0) then exit;
      end;
    result:= E_FAIL;
  end;

  //******************* TDropSource.DoEnumFormatEtc *************************
  function TDropSource.DoEnumFormatEtc(dwDirection: LongInt;
           OUT EnumFormatEtc:IEnumFormatEtc): HRESULT;
  begin
    if (dwDirection = DATADIR_GET) then
    begin
      EnumFormatEtc :=
        TEnumFormatEtc.Create(@fDataFormats, DataFormatsCount, 0);
      result := S_OK;
    end else if (dwDirection = DATADIR_SET) then
      result := E_NOTIMPL
    else result := E_INVALIDARG;
  end;

  //******************* TDropSource.GetCanonicalFormatEtc *************************
  function TDropSource.GetCanonicalFormatEtc(const FormatEtc: TFormatEtc;
           OUT FormatEtcOut: TFormatEtc): HRESULT;
  begin
    result := DATA_S_SAMEFORMATETC;
  end;

  //******************* TDropSource.SetData *************************
  function TDropSource.SetData(const FormatEtc: TFormatEtc; var Medium: TStgMedium;
           fRelease: Bool): HRESULT;
  begin
    result := E_NOTIMPL;
  end;

  //******************* TDropSource.DAdvise *************************
  function tdropsource.DAdvise(const FormatEtc: TFormatEtc; advf: LongInt;
           const advSink: IAdviseSink; OUT dwConnection: LongInt): HRESULT;
  begin
    result := OLE_E_ADVISENOTSUPPORTED;
  end;

  //******************* TDropSource.DUnadvise *************************
  function TDropSource.DUnadvise(dwConnection: LongInt): HRESULT;
  begin
    result := OLE_E_ADVISENOTSUPPORTED;
  end;

  //******************* TDropSource.EnumDAdvise *************************
  function tdropsource.EnumDAdvise(OUT EnumAdvise: IEnumStatData): HRESULT;
  begin
    result := OLE_E_ADVISENOTSUPPORTED;
  end;

  //******************* TDropSource.GetData *************************
  function tdropsource.GetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT; StdCall;
  begin
    result := DoGetData(FormatEtcIn, Medium);
  end;

  //******************* TDropSource.GetDataHere *************************
  function TDropSource.GetDataHere(const FormatEtc: TFormatEtc;
           OUT Medium: TStgMedium):HRESULT; StdCall;
  begin
    result := DoGetDataHere(FormatEtc, Medium);
  end;

  //******************* TDropSource.QueryGetData *************************
  function TDropSource.QueryGetData(const FormatEtc: TFormatEtc): HRESULT; StdCall;
  begin
    result := DoQueryGetData(FormatEtc);
  end;

  //******************* TDropSource.EnumFormatEtc *************************
  function TDropSource.EnumFormatEtc(dwDirection: LongInt;
           OUT EnumFormatEtc:IEnumFormatEtc): HRESULT; StdCall;
  begin
    result := DoEnumFormatEtc(dwDirection, EnumFormatEtc);
  end;

  //******************* TDropSource.DoGetData *************************
  function TDropSource.DoGetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT;
  begin
    result := DV_E_FORMATETC;
  end;

  //******************* TDropSource.DoGetDataHere *************************
  function TDropSource.DoGetDataHere(const FormatEtc: TFormatEtc;
           OUT Medium: TStgMedium):hresult;
  begin
    result := DV_E_FORMATETC;
  end;

  //******************* TDropSource.Hook *************************

  type
    // Dummy class used to gain access to TControls protected members
    TControlAccess = class(TControl)  
    end;

  procedure TDropSource.Hook; 
  begin
    if (FDropSource = nil) or (csDesigning in ComponentState) then
      exit;
    // Hook into the drop source's mouse handlers
    FOnMouseDown := TControlAccess(FDropSource).OnMouseDown;
    FOnMouseMove := TControlAccess(FDropSource).OnMouseMove;
    TControlAccess(FDropSource).OnMouseDown := DoOnMouseDown;
    TControlAccess(FDropSource).OnMouseMove := DoOnMouseMove;
  end;

  //******************* TDropSource.Unhook *************************
  procedure TDropSource.Unhook; 
  begin
    if (FDropSource <> nil) and (not(csDesigning in ComponentState)) then
    begin
      // Unhook the drop source's mouse handlers by restoring old handlers
      TControlAccess(FDropSource).OnMouseDown := FOnMouseDown;
      TControlAccess(FDropSource).OnMouseMove := FOnMouseMove;
    end;
    FOnMouseDown := nil;
    FOnMouseMove := nil;
  end;

  //******************* TDropSource.SetDropSource *************************
  procedure TDropSource.SetDropSource(const Value: TControl);
  begin
    if (FDropSource = Value) then
      exit;
    // Unhook old source
    Unhook;
    FDropSource := Value;
    // Hook new source
    Hook;
  end;

  //******************* TDropSource.Loaded *************************
  procedure TDropSource.Loaded; 
  begin
    inherited Loaded;
    if (Assigned(FDropSource)) and (not (csDesigning in ComponentState)) then
      Hook;
  end;

  //******************* TDropSource.Notification *************************
  procedure TDropSource.Notification(AComponent: TComponent; Operation: TOperation); 
  begin
    inherited Notification(AComponent, Operation);
    if (Operation = opRemove) and (AComponent = FDropSource) then
    begin
      Unhook;
      FDropSource := nil;
    end;
  end;

  //******************* TDropSource.DoOnMouseDown *************************
  procedure TDropSource.DoOnMouseDown(Sender: TObject; Button: TMouseButton;
    Shift: TShiftState; X, Y: Integer); 
  begin
    FDragPoint := Point(X,Y);
  end;

  //******************* TDropSource.DoOnMouseMove *************************
  procedure TDropSource.DoOnMouseMove(Sender: TObject; Shift: TShiftState;
    X, Y: Integer); 
  var
    DoDrag: boolean;
  begin
    // Drag with left or right button and...
    // ...Make sure mouse has moved at least 10 pixels first.
    if (Shift * [ssLeft, ssRight] = []) or ((abs(FDragPoint.X - X) < 10)
      and (abs(FDragPoint.Y - Y) < 10)) then
      exit;

    DoDrag := True;
    if (Assigned(FOnStartDrag)) then
      FOnStartDrag(Self, DoDrag);

    if (DoDrag) then
      Execute;
  end;

  // -----------------------------------------------------------------------------
  //			TDropTextSource
  // -----------------------------------------------------------------------------

  var
    CF_FILEGROUPDESCRIPTOR, CF_FILECONTENTS: UINT;

  //******************* TDropTextSource.Create *************************
  constructor TDropTextSource.Create(aOwner: TComponent);
  begin
    inherited Create(aOwner);
    fText := '';
    CF_FILECONTENTS := RegisterClipboardFormat(CFSTR_FILECONTENTS);
    CF_FILEGROUPDESCRIPTOR := RegisterClipboardFormat(CFSTR_FILEDESCRIPTOR);

    fDataFormats[0].cfFormat := CF_TEXT;
    fDataFormats[0].ptd := NIL;
    fDataFormats[0].dwAspect := DVASPECT_CONTENT;
    fDataFormats[0].lIndex := -1;
    fDataFormats[0].tymed := TYMED_HGLOBAL;

    fDataFormats[1].cfFormat := CF_FILEGROUPDESCRIPTOR;
    fDataFormats[1].ptd := NIL;
    fDataFormats[1].dwAspect := DVASPECT_CONTENT;
    fDataFormats[1].lIndex := -1;
    fDataFormats[1].tymed := TYMED_HGLOBAL;

    fDataFormats[2].cfFormat := CF_FILECONTENTS;
    fDataFormats[2].ptd := NIL;
    fDataFormats[2].dwAspect := DVASPECT_CONTENT;
    fDataFormats[2].lIndex := -1;
    fDataFormats[2].tymed := TYMED_HGLOBAL;

    DataFormatsCount := 3;
  end;

  // Adapted from Zbysek Hlinka, zhlinka@login.cz.
  //******************* TDropTextSource.CopyToClipboard *************************
  function TDropTextSource.CopyToClipboard: boolean;
  var
    FormatEtcIn: TFormatEtc;
    Medium: TStgMedium;
  begin
    FormatEtcIn.cfFormat := CF_TEXT;
    FormatEtcIn.dwAspect := DVASPECT_CONTENT;
    FormatEtcIn.tymed := TYMED_HGLOBAL;
    if fText = '' then result := false
    else if GetData(formatetcIn,Medium) = S_OK then
    begin
      Clipboard.SetAsHandle(CF_TEXT,Medium.hGlobal);
      result := true;
    end else result := false;
  end;

  //******************* TDropTextSource.DoGetDataHere *************************
  function TDropTextSource.DoGetDataHere(const FormatEtc: TFormatEtc; OUT Medium: TStgMedium):HRESULT;
  var
    pText: PChar;
  begin
    if (FormatEtc.cfFormat = CF_TEXT) and
      (FormatEtc.dwAspect = DVASPECT_CONTENT) and
      (FormatEtc.tymed = TYMED_HGLOBAL) and (Medium.tymed = TYMED_HGLOBAL) then
    begin
      if (Medium.hGlobal = 0) then
      begin
        result := E_OUTOFMEMORY;
        Exit;
      end;
      pText := PChar(GlobalLock(Medium.hGlobal));
      try
        StrCopy(pText, PChar(fText));
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
      Medium.UnkForRelease := NIL;
      result := S_OK;
    end else
      result := inherited DoGetDataHere(FormatEtc, Medium);
  end;

  // Adapted from stefc@fabula.com
  //******************* TDropTextSource.DoGetData *************************
  function TDropTextSource.DoGetData(const FormatEtcIn: TFormatEtc; OUT Medium: TStgMedium):HRESULT;
  var
    pFGD: PFileGroupDescriptor;
    pText: PChar;
  begin

    Medium.tymed := 0;
    Medium.UnkForRelease := NIL;
    Medium.hGlobal := 0;

    if (FormatEtcIn.cfFormat = CF_TEXT) and
      (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
      (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
    begin
      Medium.hGlobal := GlobalAlloc(GMEM_SHARE or GHND, Length(fText)+1);
      if (Medium.hGlobal = 0) then
      begin
        result := E_OUTOFMEMORY;
        Exit;
      end;
      medium.tymed := TYMED_HGLOBAL;
      result := DoGetDataHere(FormatEtcIn, Medium);
      //if RenderTextAsHGlobal(fText,Medium.hGlobal) then result := S_OK;
    end
    else if (FormatEtcIn.cfFormat = CF_FILEGROUPDESCRIPTOR) and
      (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
      (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
    begin
      Medium.hGlobal := GlobalAlloc(GMEM_SHARE or GHND, SizeOf(TFileGroupDescriptor));
      if (Medium.hGlobal = 0) then
      begin
        result := E_OUTOFMEMORY;
        Exit;
      end;
      medium.tymed := TYMED_HGLOBAL;
      pFGD := pointer(GlobalLock(Medium.hGlobal));
      with pFGD^ do
      begin
        cItems := 1;
        fgd[0].dwFlags := FD_LINKUI;
        fgd[0].cFileName := 'Text Scrap File.txt';
      end;
      GlobalUnlock(Medium.hGlobal);
      result := S_OK;
    end
    else if (FormatEtcIn.cfFormat = CF_FILECONTENTS) and
      (FormatEtcIn.dwAspect = DVASPECT_CONTENT) and
      (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
    begin
      Medium.hGlobal := GlobalAlloc(GMEM_SHARE or GHND, Length(fText)+1);
      if (Medium.hGlobal = 0) then
      begin
        result := E_OUTOFMEMORY;
        Exit;
      end;
      medium.tymed := TYMED_HGLOBAL;

      pText := PChar(GlobalLock(Medium.hGlobal));
      StrCopy(pText, PChar(fText));
      //if RenderTextAsHGlobal(fText,Medium.hGlobal) then result := S_OK;

      GlobalUnlock(Medium.hGlobal);
      result := S_OK;
    end else
      result := DV_E_FORMATETC;
  end;

  // -----------------------------------------------------------------------------
  //			TDropFileSource
  // -----------------------------------------------------------------------------

  var
    CF_IDLIST: UINT;

  //******************* TDropFileSource.Create *************************
  constructor TDropFileSource.Create(aOwner: TComponent);
  begin
    inherited Create(aOwner);
    fFiles := TStringList.Create;
    CF_IDLIST := RegisterClipboardFormat(CFSTR_SHELLIDLIST);

    fDataFormats[0].cfFormat := CF_HDROP;
    fDataFormats[0].ptd      := NIL;
    fDataFormats[0].dwAspect := DVASPECT_CONTENT;
    fDataFormats[0].lIndex   := -1;
    fDataFormats[0].tymed    := TYMED_HGLOBAL;

    fDataFormats[1].cfFormat := CF_IDLIST;
    fDataFormats[1].ptd      := NIL;
    fDataFormats[1].dwAspect := DVASPECT_CONTENT;
    fDataFormats[1].lIndex   := -1;
    fDataFormats[1].tymed    := TYMED_HGLOBAL;

    DataFormatsCount := 2;
  end;

  //******************* TDropFileSource.Destroy *************************
  destructor TDropFileSource.destroy;
  begin
    fFiles.Free;
    inherited Destroy;
  end;

  // Adapted from Zbysek Hlinka, zhlinka@login.cz.
  //******************* TDropFileSource.CopyToClipboard *************************
  function TDropFileSource.CopyToClipboard: boolean;
  var
    FormatEtcIn: TFormatEtc;
    Medium: TStgMedium;
  begin
    FormatEtcIn.cfFormat := CF_HDROP;
    FormatEtcIn.dwAspect := DVASPECT_CONTENT;
    FormatEtcIn.tymed := TYMED_HGLOBAL;
    if Files.count = 0 then result := false
    else if GetData(formatetcIn,Medium) = S_OK then
    begin
      Clipboard.SetAsHandle(CF_HDROP,Medium.hGlobal);
      result := true;
    end else result := false;
  end;

  //******************* TDropFileSource.DoGetDataHere *************************
  function TDropFileSource.DoGetDataHere(const FormatEtc: TFormatEtc;
           OUT Medium: TStgMedium):HRESULT;
  var
    i: Integer;
    dropfiles: pMyDropFiles;
    pText: PChar;
  begin
    if (FormatEtc.cfFormat = CF_HDROP) and
      ((FormatEtc.dwAspect = DVASPECT_CONTENT) or (FormatEtc.dwAspect = DVASPECT_ICON)) and
      (FormatEtc.tymed = TYMED_HGLOBAL) and (Medium.tymed = TYMED_HGLOBAL) then
    begin
      if (Medium.hGlobal = 0) then
      begin
        result:=E_OUTOFMEMORY;
        Exit;
      end;
      dropfiles := GlobalLock(Medium.hGlobal);
      try
        dropfiles^.dropfiles.pfiles := SizeOf(TDropFiles);
        dropfiles^.dropfiles.fwide := False;
        pText := @(dropfiles^.filenames);
        for i := 0 to fFiles.Count-1 DO
        begin
          StrPCopy(pText, fFiles[i]);
          Inc(pText, Length(fFiles[i])+1);
        end;
        pText^ := #0;
      finally
        GlobalUnlock(Medium.hGlobal);
      end;
      Medium.UnkForRelease := NIL;
      result := S_OK;
    end else
      result := inherited DoGetDataHere(FormatEtc, Medium);
  end;

  //******************* TDropFileSource.DoGetData *************************
  function TDropFileSource.DoGetData(const FormatEtcIn: TFormatEtc;
           OUT Medium: TStgMedium):HRESULT;
  var
    i: Integer;
    strlength: Integer;
    tmpFilenames: TStringList;
  begin
    Medium.tymed := 0;
    Medium.UnkForRelease := NIL;
    Medium.hGlobal := 0;

    if fFiles.count = 0 then result := E_UNEXPECTED
    else if (FormatEtcIn.cfFormat = CF_HDROP) and
      ((FormatEtcIn.dwAspect = DVASPECT_CONTENT) or (FormatEtcIn.dwAspect = DVASPECT_ICON)) and
      (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
    begin
      strlength := 0;
      for i := 0 to fFiles.Count-1 DO
        Inc(strlength, Length(fFiles[i])+1);
      Medium.hGlobal := GlobalAlloc(GMEM_SHARE or GMEM_ZEROINIT, SizeOf(TDropFiles)+strlength+1);
      if (Medium.hGlobal = 0) then
      begin
        result:=E_OUTOFMEMORY;
        Exit;
      end;
      Medium.tymed := TYMED_HGLOBAL;
      result := DoGetDataHere(FormatEtcIn, Medium);
      end
    else if (FormatEtcIn.cfFormat = CF_IDLIST) and
      ((FormatEtcIn.dwAspect = DVASPECT_CONTENT) or (FormatEtcIn.dwAspect = DVASPECT_ICON)) and
      (FormatEtcIn.tymed and TYMED_HGLOBAL <> 0) then
    begin
      tmpFilenames := TStringList.create;
      Medium.tymed := TYMED_HGLOBAL;
      for i := 0 to fFiles.count-1 do
        tmpFilenames.add(extractfilename(fFiles[i]));
      Medium.hGlobal :=
          ConvertFilesToShellIDList(extractfilepath(fFiles[0]),tmpFilenames);
      if Medium.hGlobal = 0 then
        result:=E_OUTOFMEMORY else
        result := S_OK;
      tmpFilenames.free;
    end else
      result := DV_E_FORMATETC;
  end;

  //********************************************
  //********************************************

initialization
  OleInitialize(NIL);
finalization
  OleUninitialize;

end.
